# These are hooks for a dumb animation state storage system, so I can
# easily call it inside a displayable's render method without the memory
# leaks that accrue when I try to misuse transforms for this and then
# roll back.

import pygame
import math
import unroller
import renpy

def anchored(anchor, span):
    if type(anchor) is float:
        return int(anchor*span)
    elif type(anchor) is int:
        return anchor
    raise Exception('Invalid anchor')
    
def eased_f(start, end, x):
    if x>1.:
        return end
    x = .5 - math.cos(math.pi*x) / 2.0
    span = end-start
    return start+(span*x)

def eased_i(start, end, x):
    if x>1:
        return end
    x = .5 - math.cos(math.pi*x) / 2.0
    span = end-start
    return start+int(span*x)

def lineared_f(start, end, x):
    if x>1.:
        return end
    span = end-start
    return start+(span*x)

def lineared_i(start, end, x):
    if x>1:
        return end
    span = end-start
    return start+int(span*x)

class AnimState(object):
    rotate = 0
    x = 0
    y = 0
    xanchor = 0
    yanchor = 0
    zoom = 1
    blend = None
    alpha = 1.0
    surf = False
    # If true, this will be used instead of the surface bassed to the func.
    # This can be used to short-circuit particularly slow operations, e.g.
    # rotozoom when the zooming is complete.
    def resolve_coords(self, surf, container):
        contw, conth = container.get_size()
        surfw, surfh = surf.get_size()
        xoff = anchored(self.xanchor, surfw)
        yoff = anchored(self.yanchor, surfh)
        return (self.x-xoff,self.y-yoff)

class AnimController(object):
    # Note that all children of AnimControllers will be loadable from saves.
    # This shouldn't be a problem unless I write an insane AnimController.
    pass

class NullAnimController(object):
    # This will be used by the pickle inspector if it can't find a valid
    # animcontroller.
    def get_state(self, t):
        return AnimState()

class TrueEndDimAnim1(AnimController):
    def get_state(self, t):
        ret = AnimState()
        ret.x = ret.y = ret.xanchor = ret.yanchor = 0
        ret.zoom = 1.0
        if t<1.0:
            ret.alpha = t*.3
        else:
            t2 = t%2.0
            if t2>=1.0:
                ret.alpha = eased_f(.33,.1,t2-1)
            else:
                ret.alpha = eased_f(.1, .33, t2)
        return ret

class TrueEndDimAnim2(AnimController):
    last_t = 0.
    def get_state(self, t):
        self.last_t = t
        ret = AnimState()
        ret.xanchor = ret.yanchor = 1.
        ret.x = 800
        ret.y = 600
        if t<=1.:
            ret.zoom = 1.0
            ret.alpha = t*.2
            return ret
        t -= 1
        if t<10.:
            ret.alpha = t*.07
        else:
            ret.alpha = .7
        ret.zoom = eased_f(1.0, 2.5, (t%2)*.5)
        return ret

class DteFlareAnim(AnimController):
    zoom = None
    move = None
    def __init__(self):
        if self.zoom is not None:
            self.zoom_start, self.zoom_end, self.zoom_dur = self.zoom
        if self.move is not None:
            self.x_start, self.y_start = self.move[0]
            self.x_end, self.y_end = self.move[1]
            self.move_dur = self.move[2]
        self.move_dur2 = self.move_dur*2
        self.zoom_dur2 = self.zoom_dur*2
    def get_state(self, t):
        ret = AnimState()
        ret.alpha = 1.0
        ret.xanchor = ret.yanchor = .5
        ret.blend = self.blend
        ret.rotate = lineared_f(0, self.rot_end, (t%self.rot_dur)/self.rot_dur)
        zoompos = t%self.zoom_dur2
        if zoompos>=self.zoom_dur:
            ret.zoom = lineared_f(self.zoom_end, self.zoom_start, (zoompos-self.zoom_dur)/self.zoom_dur)
        else:
            ret.zoom = lineared_f(self.zoom_start, self.zoom_end, zoompos/self.zoom_dur)
        movepos = t % self.move_dur2
        if movepos>=self.move_dur:
            movefrac = (movepos-self.move_dur)/self.move_dur
            ret.x = lineared_i(self.x_end, self.x_start, movefrac)
            ret.y = lineared_i(self.y_end, self.y_start, movefrac)
        else:
            ret.x = lineared_i(self.x_start, self.x_end, movepos/self.move_dur)
            ret.y = lineared_i(self.y_start, self.y_end, movepos/self.move_dur)
        return ret

class DteFlareAnim4(DteFlareAnim):
    #show flare_flat as flare4 at RotoZoom(0,-359,1.3,0.2,0.05,4.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((100,30,0.5,0.5),(500,320,0.5,0.5),3.0,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359.
    rot_dur = 1.3
    blend = pygame.BLEND_ADD
    zoom_start = .2
    zoom_end = .05
    zoom_dur = 4.3
    move = ((-12, -10),(388,280),3.)

class DteFlareAnim5(DteFlareAnim):
    #show flare_flat as flare5 at RotoZoom(0,-359,1.7,0.3,0.1,3.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_SUB),Move((100,230,0.5,0.5),(490,330,0.5,0.5),2.7,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359.
    rot_dur = 1.7
    zoom_start = .3
    zoom_end = .1
    zoom_dur = 3.3
    blend = pygame.BLEND_SUB
    move = ((-12,190),(378,290),2.7)

class DteFlareAnim3(DteFlareAnim):
    #show flare_flat as flare3 at RotoZoom(0,-359,1.3,0.6,0.3,2.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((1124,400,0.5,0.5),(544,292,0.5,0.5),7.0,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359.
    rot_dur = 1.3
    zoom_start = .6
    zoom_end = .3
    zoom_dur = 2.3
    blend = pygame.BLEND_ADD
    move = ((1020,360),(432,252),7.)

class DteFlareAnim6(DteFlareAnim):
    #show flare_flat as flare6 at RotoZoom(0,-359,1.3,0.25,0.15,4.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((40,640,0.5,0.5),(588,257,0.5,0.5),5.0,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359.
    rot_dur = 1.3
    zoom = (.25, .15, 4.3)
    #move = ((144, 614), (692, 231), 5.0)
    move = ((-72, 600), (476, 217), 5.)
    blend = pygame.BLEND_ADD

class DteFlareAnimEne(DteFlareAnim):
    #show flare_flat as flarene at RotoZoom(0,-359,1.3,0.7,0.2,4.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_SUB),Move((912,-20,0.5,0.5),(429,383,0.5,0.5),8.0,repeat=True,bounce=True) behind vn_overlaya
    rot_end = -359
    rot_dur = 1.3
    zoom = (.7,.2,4.3)
    blend = pygame.BLEND_SUB
    move = ((800,-60),(317,343),8.)

class DteFlareAnimSe(DteFlareAnim):
    rot_end = -359
    rot_dur = 1.3
    zoom = (.65, .15,4.3)
    blend = pygame.BLEND_ADD
    move = ((800,780),(382,273),6.9)
    #show flare_flat as flarese at RotoZoom(0,-359,1.3,0.65,0.15,4.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((912,820,0.5,0.5),(494,313,0.5,0.5),6.9,repeat=True,bounce=True) behind vn_overlay

class DteFlareAnimN(DteFlareAnim):
    #show flare_flat as flare_n at RotoZoom(0,-359,1.4,0.25,0.10,3.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((507,14,0.5,0.5),(593,207,0.5,0.5),4.9,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359
    rot_dur = 1.4
    zoom = (.25, .1, 3.3)
    blend = pygame.BLEND_ADD
    move = ((395,-26), (481, 167), 4.9)

class DteFlareAnimS(DteFlareAnim):
    #show flare_flat as flares at RotoZoom(0,-359,1.3,0.95,0.45,3.3,rot_repeat=True,zoom_bounce=True,zoom_repeat=True),Blender(pygame.BLEND_ADD),Move((600,820,0.5,0.5),(503,370,0.5,0.5),8.9,repeat=True,bounce=True) behind vn_overlay
    rot_end = -359
    rot_dur = 1.3
    zoom = (.95,.45,3.3)
    blend = pygame.BLEND_ADD
    move = ((488,780),(391,330),8.9)

class DteBgShake(AnimController):
    #show cg lab_floating as bg:
    #    pos (112,40)
    #    parallel:
    #        linear 0.2 pos (90,30)
    #        linear 0.4 pos (134,50)
    #        linear 0.2 pos (112,40)
    #        repeat
    def get_state(self,t):
        ret = AnimState()
        ret.alpha = 1.0
        t = t%.8
        if t<.2:
            # -22 was too far, left a bit of black at the edge.
            ret.x = lineared_i(0,-18,t*5.) 
            ret.y = lineared_i(0,-10,t*5.)
        elif t<.6:
            t -= .2
            ret.x = lineared_i(-18, 22, t*2.5)
            ret.y = lineared_i(-10, 10, t*2.5)
        else:
            t -= .6
            ret.x = lineared_i(22, 0, t*5.)
            ret.y = lineared_i(10, 0, t*5.)
        return ret

class DteDim2FadeDown(AnimController):
    startalpha = .4
    zoomdelt = 0.
    def __init__(self, take=None):
        if take is not None:
            self.zoomdelt = take.last_t
            data = take.get_state(take.last_t)
            self.startalpha = data.alpha
    def get_state(self, t):
        ret = AnimState()
        ret.xanchor = ret.yanchor = 1.
        ret.x = 800
        ret.y = 600
        ret.alpha = eased_f(self.startalpha, .4, t*.1)
        ret.zoom = eased_f(1.0,2.5,((t-self.zoomdelt)%2)*.5)
        return ret

class DteBgShakeFadeIn1_5(DteBgShake):
    def get_state(self, t):
        ret = DteBgShake.get_state(self, t)
        ret.alpha = lineared_f(0.,1.,t/1.5)
        return ret

def centeredrect(surf, wh):
    neww, newh = wh
    oldw, oldh = surf.get_size()
    if oldw>neww:
        x = (oldw-neww)>>1
        w = neww
    else:
        x = 0
        w = oldw
    if oldh>newh:
        y = (oldh-newh)>>1
        h = newh
    else:
        h = oldh
        y = 0
    return surf.subsurface((x,y,w,h)).convert()

class DteSubCcw(AnimController):
    def get_state(self, t):
        ret = AnimState()
        ret.rotate = lineared_f(0.,-359.,(t%2.)*.5)
        ret.zoom = lineared_f(0.05, 4., t*.2)
        ret.blend = pygame.BLEND_SUB
        ret.xanchor = ret.yanchor = .5
        ret.x = 439
        ret.y = 239
        if ret.zoom>=4.0:
            surf = unroller.surfcache.get('DteSubCcw',None)
            if surf is None:
                # The actual range is ~460, ~378 from the centers to the edge,
                # so let's trim to (920, 760)
                unroller.surfcache['DteSubCcw'] = surf = centeredrect(renpy.display.im.cache.get(renpy.display.im.FactorScale(renpy.display.im.MatrixColor('cg/flare-flat.png',renpy.display.im.matrix.brightness(-0.2)),4.0)),(920,920))
            ret.zoom = 1.0
            ret.surf =surf
        return ret
        #show flare_flat_dim as sub_ccw behind vn_overlay at RotoZoom(0,-359,2.0,0.05,4.0,5.0,rot_repeat=True),Blender(pygame.BLEND_SUB),Position(xpos=551,ypos=279,xanchor=0.5,yanchor=0.5)

class DteSubCw(AnimController):
    def get_state(self, t):
        #show flare_flat_dim as sub_cw behind vn_overlay at RotoZoom(0,359,2.0,0.05,3.5,5.0,rot_repeat=True),Blender(pygame.BLEND_ADD),Position(xpos=561,ypos=267,xanchor=0.5,yanchor=0.5)
        ret = AnimState()
        ret.rotate = lineared_f(0.,359.,(t%2.)*.5)
        ret.zoom = lineared_f(0.05, 3.5, t*.2)
        ret.blend = pygame.BLEND_ADD
        ret.xanchor = ret.yanchor = .5
        ret.x = 449
        ret.y = 227
        if ret.zoom>=3.5:
            surf = unroller.surfcache.get('DteSubCw',None)
            if surf is None:
                unroller.surfcache['DteSubCw'] = surf = centeredrect(renpy.display.im.cache.get(renpy.display.im.FactorScale(renpy.display.im.MatrixColor('cg/flare-flat.png',renpy.display.im.matrix.brightness(-0.2)),3.5)),(920,920))
            ret.zoom = 1.0
            ret.surf = surf
        return ret
