# nopickles.py is copyright 2014 Spiky Caterpillar and released under the
# WTFPL.

# These are the rather hackish routines to replace Ren'Py's use of Python
# pickles.  Note that while this is a lot more secure than pickle, it's not
# currently fully secure.  In particular, note that it ships with
# INSECURE=True set.

# Before you panic, note that the only way I know of to actually exploit
# the pickles is for someone to have write access to your saved games or
# persistent data - so things like trading saves and persistent, or letting
# someone/something else write your files, or using networked providers
# to store your saves.  So it's a real risk, but only a HUGE one if I were
# using some form of networked save system.

# Actually using this requires some minor changes to Ren'Py (diffing the
# Date Warp Ren'Py against a stock 6.17.7 should tell you what these changes
# are) and constraining your code to only the syntax that the nopickles
# inspector handles.  If you're seriously planning on using it, you may want
# to talk it over with Spiky for integration advice.

import renpy
import unroller
import os
from namedtransform import *
import sanitize
import animation

INSECURE = True
DEBUG = True

# If set, autosave on every single line for debugging purposes.
AUTOHAMMER = False

FUTUREDEBUG = False # Set to True to provide debug information about things
# which are not a problem for Date Warp but may be a problem for future games.

# Warnings are a list of sequences; the first item of the sequence is the
# severity, the second and subsequent items are data.
#
# Severities:
# 5: Discarded data.
# 6: Accepted tainted data.
class Warnings(list):
    def __init__(self):
        list.__init__(self)
        self.notes = []
    def note(self,v):
        self.notes.append(v)
    def unnote(self):
        self.notes.pop()
    def append(self, v):
        list.append(self,v+tuple(self.notes))

def clear_warnings():
    global warnings
    warnings = Warnings()

clear_warnings()

def most_important_warning():
    highest = -1
    for warning in warnings:
        if warning[0]>highest:
            highest = warning[0]
            highest_val = warning
    if highest>=0:
        ret = repr(highest_val)
        return ret.replace('{','{{')
    return '(No warnings.)'

def serialize_dict_contents(item):
    ret = []
    for k, v in item.iteritems():
        ret.extend(serialize_value(k))
        ret.extend(serialize_value(v))
    return ret

def serialize_value(item):
    ti = type(item)
    if ti is tuple:
        ret = ['(',]
        for i in item:
            ret.extend(serialize_value(i))
        ret.append(')')
        return ret
    elif item is None:
        return 'None',
    elif ti is str or ti is unicode:
        return 'str '+escape_str(item),
    elif ti is int:
        return 'int '+str(item),
    elif ti is float:
        return 'float '+str(item),
    elif ti is bool:
        return str(item),
    elif ti is dict:
        ret = ['p{',]+serialize_dict_contents(item)+['}',]
        return ret
    elif ti is renpy.python.RevertableDict:
        ret = ['{']
        ret.extend(serialize_dict_contents(item))
        return ret + ['}',]
    elif ti is renpy.python.RevertableList:
        ret = ['[',]
        for i in item:
            ret.extend(serialize_value(i))
        return ret + [']',]
    elif ti is NamedTransform:
       return ('trans '+item.name,)
    elif animation.AnimController in ti.mro():
        return ('ac '+ti.__name__,)
    print "BUG: Cannot serialize values of type",type(item)
    return ('#.44 unhandled value'+str(type(item)),)

def serialize_tnv(k, v):
    # Serialize a key/value pair to a sequence of one or more strings.
    # The format is type, name, value.
    # Non-scalar values will extend over multiple strings in the returned
    # sequence.
    if type(v) is str or type(v) is unicode:
        return 'str '+k+' '+escape_str(v),
    elif type(v) is int:
        return 'int '+k+' '+str(v),
    elif type(v) is bool:
        return str(v)+' '+k,
    elif type(v) is float:
        return 'float '+k+' '+str(v),
    elif type(v) is tuple:
        ret = ['( '+k]
        for i in v:
            ret.extend(serialize_value(i))
        ret.append(')')
        return ret
    elif v is None:
        return 'None '+k,
    elif type(v) is renpy.python.RevertableDict:
        ret = ['{ '+k]
        for dk, dv in v.iteritems():
            ret.extend(serialize_value(dk))
            ret.extend(serialize_value(dv))
        ret.append('}')
        return ret
    elif type(v) is dict:
        ret = ['p{ '+k]
        for dk, dv in v.iteritems():
            ret.extend(serialize_value(dk))
            ret.extend(serialize_value(dv))
        ret.append('}')
        return ret
    elif type(v) is renpy.python.RevertableList:
        ret = ['[ '+k]
        for i in v:
            ret.extend(serialize_value(i))
        ret.append(']')
        return ret
    elif type(v) is renpy.python.RevertableSet:
        ret = ['rs( '+k]
        for i in v:
            ret.extend(serialize_value(i))
        ret.append(')')
        return ret
    elif type(v) is renpy.python.StoreDeleted:
        return ('del '+k,)
    else:
        return '# Invalid tnv: '+k+':'+str(type(v)),
        
def serialize_script_loc(loc):
    # We reverse the order of locations so they're line, timestamp, file
    # so that files with spaces will not screw up parsing.

    return str(loc[2])+' '+str(loc[1])+' '+loc[0]

def serialize_rollback_log(log):
    roots = renpy.game.log.freeze()
    #roots = log.get_roots()
    ret = ['log',]
    found = False
    l = list(log.log)
    l.reverse()
    elongatable = []
    for ent in l:
        #ret.append("#.156: "+str(roots))
        es = SerializedRollbackLogEntry(ent, roots)
        if es.can_be_last():
            roots = False
            found = True
        elongatable.append(es)
    elongatable.reverse()
    for es in elongatable:
        ret.extend(es.elongate())
    ret.append('end')
    if not found:
        raise Exception("No valid serialization points found.")
    return ret

def serialize_context_musicchannel(channelname, data):
    ret = []
    if data.last_filenames:
        for fn in data.last_filenames:
            ret.append('file '+escape_str(fn))
        if data.secondary_volume!=1.0:
            ret.append('vol '+str(data.secondary_volume))
        if data.pan!=0:
            print "Warning: nopickles does not currently save pan values."
    if ret:
        return ['channel '+str(channelname),]+ret+['end',]
    return ()

def serialize_context_music(music):
    # A context's music is a Ren'Py RevertableDict.  I'm not entirely sure why.
    ret = []
    for k,v in music.iteritems():
        ret.extend(serialize_context_musicchannel(k,v))
    if ret:
        return ['music',]+ret+['end',]
    return ()

ROLLBACK_OBJECT_TYPECODES={
    renpy.python.RevertableObject:'ro'
    }

def serialize_rollback_object(obj, data):
    to = type(obj)
    if to not in ROLLBACK_OBJECT_TYPECODES:
        return '#.155'+str(type(obj)),
    ret = ["%x"%id(obj)+" "+ROLLBACK_OBJECT_TYPECODES[to],]
    return ret+serialize_value(data)


class SerializedRollbackLogEntry(object):
    location_makes_sense = False
    has_roots = False
    def __init__(self, entry, roots):
        self.renpy_entry = entry
        cloc = entry.context.current
        # Only return members of the rollback log that make sense - experimental
        if type(cloc) is not tuple:
            return
        if len(cloc) != 3:
            return
        if type(cloc[2]) is not int or type(cloc[1]) is not int or type(cloc[0]) is not unicode:
            return
        self.location_makes_sense = True
        self.data = ret = ['entry','callstack']
        for loc in entry.context.call_location_stack:
            ret.append(serialize_script_loc(loc))
        ret.append('end')
        ret.extend(serialize_context_music(entry.context.music))
        for item in entry.context.dynamic_stack:
            # These are regular dicts.
            ret.append('dynamic')
            for k,v in item.iteritems():
                ret.extend(serialize_tnv(k,v))
            ret.append('end')
        ret.append('current '+serialize_script_loc(cloc))
        ret.append('#168.')
        for storename, storeval in entry.stores.iteritems():
            if storeval:
                ret.append('store '+storename)
                for k,v in storeval.iteritems():
                    ret.extend(serialize_tnv(k,v))
                ret.append('end')
        # Behind experimental rollback-related saving.
        #if entry.objects:
        #    ret.append('objects')
        #    for obj, data in entry.objects:
        #       ret.extend(serialize_rollback_object(obj, data))
        #    ret.append('end')
        # End experimental rollback-related saving.
        self.root_chunk = ['#.root chunk']
        try:
            if roots:
                objmemo = {}
                for k,v in entry.objects:
                    objmemo[id(k)] = v
                for k,v in roots.iteritems():
                    #try:
                    #    print "Debug:",k,v
                    #except Exception,e:
                    #    print "Exception printing debug message:",e
                    if id(v) in objmemo:
                        self.root_chunk.append('#.211 rollback '+str(v)+'->'+str(objmemo[id(v)]))
                        nv = type(v)()
                        nv.rollback(objmemo[id(v)])
                        self.root_chunk.extend(serialize_tnv(k,nv))
                        #self.root_chunk.append('r '+k)
                        #self.root_chunk.extend(serialize_value(objmemo[id(v)]))
                        #self.root_chunk.append('/r')
                    else:
                        self.root_chunk.extend(serialize_tnv(k,v))
            self.has_roots = True
            self.root_chunk.append('#/root chunk')
        except Exception,e:
            self.root_exception = e
        ret.extend(serialize_scene_lists(entry.context.scene_lists))
    def can_be_last(self):
        return self.location_makes_sense and self.has_roots
    def objects_chunk(self):
        ret = []
        if FUTUREDEBUG:
            for obj in self.renpy_entry.objects:
                ret.append('#.217'+str(obj))
        return ret
    def elongate(self):
        if not self.location_makes_sense:
            return ()
        if self.has_roots:
            return self.data + self.root_chunk + self.objects_chunk()+ ['end',]
        return self.data+['end',]

alphalow = 'abcdefghijklmnopqrstuvwxyz'

def clean_string(s, allow, replace='?'):
    ret = ''
    for c in s:
        if c in allow:
            ret += c
        else:
            ret += replace
    return ret

def unserialize_value(context_stack, line):
    if line=='(':
        context_stack.append(SafeloadTupleValueContext())
        return context_stack[-1]
    elif line=='[':
        context_stack.append(SafeloadRevertableListValueContext())
        return context_stack[-1]
    elif line=='p{':
        context_stack.append(SafeloadPythonDictValueContext())
        return context_stack[-1]
    elif line=='{':
        context_stack.append(SafeloadRevertableDictValueContext())
        return context_stack[-1]
    elif line=='False':
        return False
    elif line=='None':
        return None
    elif line=='True':
        return True
    elif line.startswith('float '):
        return float(line[6:])
    elif line.startswith('int '):
        return int(line[4:])
    elif line.startswith('str '):
        return unescape_str(line[4:])
    elif line.startswith('trans '):
        trans = getattr(renpy.store,line[6:],None)
        if type(trans) is NamedTransform:
            return trans
        else:
            warnings.append((5,"unserialize_value discarded missing transform",trans))
            ret = getattr(renpy.store,'unresolvable')
            if type(ret) is not NamedTransform:
                raise Exception("renpy.store.default is not a NamedTransform.")
            return ret
    elif line.startswith('ac '):
        candidate = getattr(animation,line[3:])
        if type(candidate) is type(animation.AnimController) and animation.AnimController in candidate.mro():
            return candidate()
        warnings.append((5,"unserialize_value used NullAnimController for",line))
        return animation.NullAnimController()
        
    raise Exception("Debug: unserialize_value discarded '"+line+"'")
    return None

class DeindirectableContext(object):
    """
    Used to check to see whether something should be deindirected before use.
    """
    pass

class SafeloadSequenceValueContext(DeindirectableContext):
    """
    This isn't meant to be used directly; rather, it's a superclass for
    tuples and lists.
    """
    def __init__(self):
        self.l = []

    def deindirect_values(self):
        i = 0
        while i<len(self.l):
            if isinstance(self.l[i], SafeloadSequenceValueContext):
                self.l[i] = self.l[i].deindirect()
            if type(self.l[i]) is SafeloadPythonDictValueContext:
                self.l[i] = self.l[i].deindirect()
            i += 1

class SafeloadRevertableListValueContext(SafeloadSequenceValueContext):
    def handle(self,context_stack,line):
        if line==']':
            context_stack.pop()
            self.deindirect_values()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        return renpy.python.RevertableList(self.l)

class SafeloadPythonDictValueContext(DeindirectableContext):
    def __init__(self):
        self.l = []
    def handle(self, context_stack, line):
        if line=='}':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def build_dict(self):
        i = 0
        d = {}
        while i<len(self.l):
            d[deindirect_if_needed(self.l[i])] = deindirect_if_needed(self.l[i+1])
            i += 2
        return d
    def deindirect(self):
        return self.build_dict()

class SafeloadRevertableDictValueContext(SafeloadPythonDictValueContext):
    def build_dict(self):
        raise Exception("Wrong type.")
    def build_revertable_dict(self):
        i = 0
        d = renpy.python.RevertableDict()
        while i<len(self.l):
            d[self.l[i]] = self.l[i+1]
            i += 2
        return d
    def deindirect(self):
        return self.build_revertable_dict()

DIRECT_TYPES = (bool, float, int, type(None), str, unicode, tuple)

def deindirect_if_needed(v):
    if isinstance(v, DeindirectableContext):
        return v.deindirect()
    if type(v) not in DIRECT_TYPES:
        if isinstance(v, animation.AnimController):
            # AnimControllers should be OK to pull out directly because
            # they're written under the assumption that they'll be loaded.
            return v
        raise Exception("Unexpected type "+str(type(v)))
    return v

class SafeloadTupleValueContext(SafeloadSequenceValueContext):
    def __init__(self):
        self.l = []
    def handle(self, context_stack, line):
        if line == ')':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        ret = tuple([deindirect_if_needed(i) for i in self.l])
        return ret

class SafeloadRevertableSetValueContext(SafeloadSequenceValueContext):
    def handle(self, context_stack, line):
        if line==')':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        return renpy.python.RevertableSet([deindirect_if_needed(i) for i in self.l])

def unserialize_script_loc(line):
    loc = line.split(' ',2)
    # XXX Locations are not sanity-checked, so a save can jump to any
    # label/line in your game _AND_ in the stock Ren'Py library.  This is
    # one of the reasons I'm not calling this 'secure' just yet.
    return (loc[2], int(loc[1]), int(loc[0]))

class SafeloadCallstackContext(object):
    def __init__(self):
        self.stack = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        self.stack.append(unserialize_script_loc(line))

def scrub_g_points(s):
    ret = renpy.python.RevertableDict()
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if type(s) is not renpy.python.RevertableDict:
        warnings.append((5,"Expected a RevertableDict, got",type(s)))
        return ret
    for k,v in s.iteritems():
        if type(k) is str and type(v) is int:
            ret[k] = v
    return ret

STORE_SCRUBBERS = dict(
    args = sanitize.expect_del_or_emptytuple,
    current_save = sanitize.discard,
    dimension_anim_base_profiles = sanitize.scrub_dimension_anim_base_profiles,
    dimension_anim_true_end_profiles = sanitize.scrub_dimension_anim_base_profiles,
    favourite = sanitize.expect_str_or_None,
    favourite_points = sanitize.expect_del_or_int,
    first_boy_followed = sanitize.expect_del_or_str,
    g_accepted_linds_offer = sanitize.expect_del_or_str,
    g_ask_rafael_about = sanitize.expect_del_or_str,
    g_encourage_bradley_to_leave = sanitize.expect_del_or_str,
    g_helped_rafael_do_dishes = sanitize.expect_bool_or_del,
    g_how_should_janet_stroke_linds = sanitize.expect_del_or_str,
    g_janet_has_never = sanitize.expect_del_or_str,
    #g_last_line = sanitize.expect_displayable_text, - No longer used.
    g_make_alben_feel_better = sanitize.expect_del_or_str,
    g_nathaniel_fail = sanitize.expect_del_or_str,
    g_obey_linds = sanitize.expect_del_or_str,
    g_points = scrub_g_points,
    g_reaction_to_alben_cheating = sanitize.expect_del_or_str,
    g_reaction_to_being_caught_with_nathaniel = sanitize.expect_del_or_str,
    g_reaction_to_being_grabbed_by_alben = sanitize.expect_del_or_str,
    g_religion_convo = sanitize.expect_del_or_str,
    g_respond_to_albens_complaints = sanitize.expect_del_or_str,
    g_reveal_bianca = sanitize.expect_del_or_str,
    g_shared_room_with_bradley = sanitize.expect_bool_or_del,
    g_stay_with_linds_and_rafael = sanitize.expect_del_or_str,
    g_what_is_janet_struggling_to_remember = sanitize.expect_del_or_str,
    gray_girl_anim_phase = sanitize.expect_del_or_int,
    jukebox_pos = sanitize.expect_int,
    jumped_between_bradley_and_alben = sanitize.expect_bool_or_del,
    keyboard_target_widget = sanitize.expect_del,
    kwargs = sanitize.expect_del_or_emptydict,
    last_timer = sanitize.expect_del_or_str,
    listened_to_bradley = sanitize.expect_bool_or_del,
    load_count = sanitize.expect_del_or_int_or_None,
    message = sanitize.expect_del_or_str, # Used by traceback uploader
    result = sanitize.expect_del_or_str, # Used by traceback uploader
    mouse_visible = sanitize.expect_bool,
    nvl_list = sanitize.scrub_nvl_list,
    old_key_repeat = sanitize.expect_del_or_int2,
    old_load_count = sanitize.expect_del_or_int,
    predict_screens = sanitize.new_empty_revertable_list,
    ret = sanitize.expect_del_or_str,
    save_caption_widget = sanitize.expect_del,
    save_name = sanitize.expect_str,
    sg_lr = sanitize.scrub_spritegroup,
    sg_ab = sanitize.scrub_spritegroup,
    sg_nab = sanitize.scrub_spritegroup,
    shardid = sanitize.expect_char_or_del,
    # Used for the Date Warp menu buttons.  Obsolete.
    #show_overlay = sanitize.expect_bool,
    show_side_buttons = sanitize.expect_bool,
    side_buttons_fading_in = sanitize.expect_bool,
    speaking = sanitize.expect_str_or_None,
    suppress_overlay = sanitize.expect_bool, # Used internally by Ren'Py.
    warnstr = sanitize.discard,
    _args = sanitize.expect_None,
    _game_menu_screen = sanitize.scrub__game_menu_screen,
    _kwargs = sanitize.expect_None,
    _last_raw_what = sanitize.expect_displayable_text_or_del_or_None,
    _last_say_what = sanitize.expect_displayable_text_or_None,
    _last_say_who = sanitize.scrub_last_say_who,
    _m1_developer__missing_scene = sanitize.expect_bool,
    _old_game_menu_screen = sanitize.scrub__game_menu_screen,
    _old_predict_screens = sanitize.new_empty_revertable_list,
    _predict_screens = sanitize.new_empty_revertable_list,
    _return = sanitize.expect_None,
    _side_image_attributes = sanitize.expect_None,
    _windows_hidden = sanitize.expect_bool,
    )
OK_VARNAME_SUBSEQUENT = set('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')
OK_VARNAME_FIRST = set('ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')

def scrub_varname(n):
    if n.startswith('__'):
        raise Exception("No replacing magic variables!")
    if n[0] not in OK_VARNAME_FIRST:
        raise Exception("Invalid character in variable name: "+n[0])
    for c in n[1:]:
        if c not in OK_VARNAME_SUBSEQUENT:
            raise Exception("Invalid character in variable name: "+c)
    return n

def scrub_tnv(line, scrubbers):
    tnv = line.split(' ',2)
    t = tnv[0]
    n = scrub_varname(tnv[1])
    # Singleton type/values first.
    if t=='None':
        rv = None
    elif t=='True':
        rv = True
    elif t=='False':
        rv = False
    elif t=='del':
        rv = renpy.python.deleted
    elif len(tnv)!=3:
        print "Debug 344:",line
        raise Exception("Malformed tnv line.")
    elif t=='str':
        rv = unescape_str(tnv[2])
    else:
        print "Flawed line was",line
        raise Exception("Unexpected tnv line.")
    if n in scrubbers:
        warnings.note('tnv '+n)
        ret = n, scrubbers[n](rv)
        warnings.unnote()
        return ret
    elif INSECURE:
        print "Debug: returning un-scrubbed entry for",line
        return n, rv
    print "No scrubber for",line
    raise Exception("No scrubber found.")

class SafeloadDynamicStackEntryContext(object):
    def __init__(self):
        self.stack = []
    def push(self):
        self.stack.append({})
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        warnings.note('dynamicstack')
        k, v = scrub_tnv(line, STORE_SCRUBBERS)
        warnings.unnote()
        self.stack[-1][k] = v

def unescape_music_filename(fn):
    unescaped = sanitize.scrub_relative_path(unescape_str(fn))
    if not unescaped.endswith('.ogg'):
        raise Exception('Bad music filename.')
    return unescaped

class SafeloadMusicChannelContext(object):
    secondary_volume = 1.0
    def __init__(self, name):
        for c in name:
            if c not in 'abcdefghijklmnopqrstuvwxyz0123456789':
                raise Exception("Disallowed characters in music channel name.")
        try: # We try integer channel names first.
            self.name = int(name)
        except:
            self.name = name
        self.filenames = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('file '):
            self.filenames.append(unescape_music_filename(line[5:]))
        elif line.startswith('vol '):
            self.secondary_volume = float(line[4:])
        else:
            raise Exception("Unexpected line in music context.")

def scrub_store_identifier(s):
    if s=='store':
        return 'store'
    raise Exception("Unexpected store identifier "+s)

class SafeloadMusicContext(object):
    def __init__(self):
        self.channels = {}
    def handle(self, context_stack, line):
        if line.startswith('channel '):
            chan = SafeloadMusicChannelContext(line[8:])
            self.channels[chan.name] = chan
            context_stack.append(chan)
        elif line=='end':
            context_stack.pop()
        else:
            print "Line was",line
            raise Exception("Unexpected line in music context.")

class Tainted(object):
    def __init__(self, val, name=None):
        self.name = name
        self.val = val
    def deindirect_val(self):
        v = self.val
        if isinstance(v,DeindirectableContext):
            return v.deindirect()
        return v

def parse_tainted_tnv(line, context_stack):
    # This returns a tainted variable containing the line's value.
    # For values that create a new parsing context, the value will be the
    # parsing context.
    tnv = line.split(' ',2)
    t = tnv[0]
    n = tnv[1]
    if len(tnv)==3:
        v = tnv[2]
        if t=='int':
            return Tainted(int(v),n)
        elif t=='str':
            return Tainted(unescape_str(v),n)
        elif t=='float':
            return Tainted(float(v),n)
    elif len(tnv)==2:
        if t=='False':
            return Tainted(False, n)
        elif t=='None':
            return Tainted(None, n)
        elif t=='True':
            return Tainted(True, n)
        elif t=='(':
            ctx = SafeloadTupleValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='[':
            ctx = SafeloadRevertableListValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='{':
            ctx = SafeloadRevertableDictValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='p{':
            ctx = SafeloadPythonDictValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='rs(':
            context_stack.append(SafeloadRevertableSetValueContext())
            return Tainted(context_stack[-1],n)
        elif t=='del':
            return Tainted(renpy.python.deleted, n)
    raise Exception('Unhandled tnv '+line)

class SafeloadStoreContext(object):
    def __init__(self):
        self.data = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        self.data.append(parse_tainted_tnv(line, context_stack))

class SafeloadObjectsContext(object):
    onval = False
    def __init__(self):
        self.trios = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif self.onval:
            self.trios[-1].append(unserialize_value(context_stack, line))
            self.onval = False
            return
        self.trios.append(line.split(' ',1))
        self.onval = True
        

class SafeloadRollbackLogEntryContext(object):
    scene = None
    dynamicstack = None
    music = None
    def __init__(self):
        #self.roots = {}
        #self.complex_roots = {}
        self.stores = {}
        self.tainted_roots = []
    def resolve_tainted_roots(self):
        self.untainted_roots = {}
        for t in self.tainted_roots:
            if t.name.startswith('store.'):
                varname = t.name[6:]
            else:
                raise Exception("Unexpected store variable name.")
            if varname in STORE_SCRUBBERS:
                warnings.note(('tr',varname))
                self.untainted_roots[t.name] = STORE_SCRUBBERS[varname](t.deindirect_val())
                warnings.unnote()
            elif INSECURE:
                warnings.append((6,"Accepted non-scrubbable variable",varname,t.deindirect_val()))
                self.untainted_roots[t.name] = t.deindirect_val()
            else:
                raise Exception("No scrubber for "+varname)
    def handle(self, context_stack, line):
        if line=='end':
            self.resolve_tainted_roots()
            """
            for varname, val in self.complex_roots.iteritems():
                if type(val) is SafeloadPythonDictValueContext:
                    self.roots[varname] = val.build_dict()
                elif type(val) is SafeloadRevertableDictValueContext:
                    self.roots[varname] = val.build_revertable_dict()
                elif type(val) is SafeloadRevertableListValueContext:
                    self.roots[varname] = renpy.python.RevertableList(val.l)
                elif type(val) is SafeloadTupleValueContext:
                    self.roots[varname] = tuple(val.l)
                else:
                    raise Exception("Unexpected complex root type.")
            self.complex_roots = {}
            """
            context_stack.pop()
            return
        elif line=='callstack':
            context_stack.append(SafeloadCallstackContext())
            self.callstack = context_stack[-1]
            return
        elif line.startswith('current '):
            self.current = unserialize_script_loc(line[8:])
            return
        elif line=='dynamic':
            if self.dynamicstack is None:
                self.dynamicstack = SafeloadDynamicStackEntryContext()
            self.dynamicstack.push()
            context_stack.append(self.dynamicstack)
            return
        elif line=='music':
            self.music = SafeloadMusicContext()
            context_stack.append(self.music)
            return
        #elif line=='objects':
        #    self.objects = SafeloadObjectsContext()
        #    context_stack.append(self.objects)
        #    return
        elif line=='scene':
            if self.scene is not None:
                raise Exception("RollbackLogEntry already had a scene.")
            context_stack.append(SafeloadSceneContext())
            self.scene = context_stack[-1]
            return
        elif line.startswith('store '):
            storename = scrub_store_identifier(line[6:])
            context_stack.append(SafeloadStoreContext())
            self.stores[storename] = context_stack[-1]
            return
        self.tainted_roots.append(parse_tainted_tnv(line, context_stack))

    OK_MUSIC_CHANNELS = set((7,'music2',))
    def music_update_data(self):
        if not self.music:
            return ()
        global last_music_update_data
        ret = {}
        for channelname, data in self.music.channels.iteritems():
            if channelname in self.OK_MUSIC_CHANNELS:
                ctx = renpy.audio.audio.MusicContext()
                ctx.last_filenames = data.filenames
                ctx.secondary_volume = data.secondary_volume
                ctx.last_changed = renpy.audio.audio.get_serial()
                ret[channelname] = ctx
            else:
                raise Exception("Invalid channel in save.")
            last_music_update_data = ret
        return ret
    def stores_update_data(self):
        ret = {}
        for storek, storev in self.stores.iteritems():
            if storek!='store':
                raise Exception("Cannot store stores other than store.")
            ret[storek] = r = dict()
            for item in storev.data:
                n = item.name
                if n in STORE_SCRUBBERS:
                    warnings.note('stores update '+n)
                    r[n] = STORE_SCRUBBERS[n](item.deindirect_val())
                    warnings.unnote()
                elif INSECURE:
                    warnings.append((6,"No scrubber for",n,item.val))
                    r[n] = item.deindirect_val()
                else:
                    print "Error",item,name,item.val
                    raise Exception("Could not find a scrubber for a store item.")
        return ret

class SafeloadRollbackLogContext(object):
    def __init__(self):
        self.entries = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line=='entry':
            context_stack.append(SafeloadRollbackLogEntryContext())
            self.entries.append(context_stack[-1])
        else:
            raise Exception("Unexpected line in RollbackLog context.")

# The loader works on a context stack.  Context zero is the outermost.

class SafeloadContextZero(object):
    scene = None
    log = None
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line=='log':
            if self.log is not None:
                raise Exception("Context zero already had a log.")
            context_stack.append(SafeloadRollbackLogContext())
            self.log = context_stack[-1]
        elif line=='saveinfo':
            context_stack.append(SaveInfo())
            self.info = context_stack[-1]
        elif line:
            print "Context zero rejecting line '"+line+"'"
            raise Exception("Inconsistent save.")
        # We ignore blank lines, to make savegame editing less annoying.

class SafeloadSceneContext(object):
    def __init__(ctx):
        ctx.layers = []
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('layer '):
            context_stack.append(SafeloadLayerContext(line[6:]))
            ctx.layers.append(context_stack[-1])
        else:
            print ctx,'got',line
            raise Exception("Inconsistent save.")
    def SceneLists(self):
        showing = renpy.display.image.ShownImageInfo()
        ret = renpy.display.core.SceneLists(None,showing)
        for layer in self.layers:
            for img in layer.images:
                if type(img) is not SafeloadScreenContext:
                    # For now, we silently discard saved screens because their
                    # scopes are problematic.  This works for Date Warp, but
                    # isn't future-proof.
                    if img.ats:
                        at_list = img.ats
                    else:
                        at_list = []
                    if img.tag is None:
                        key = img.what
                    else:
                        key = img.tag
                    if img.tag.startswith('hide$'):
                        warnings.append((0,"Save contained a hiding displayable; discarded it."))
                        continue
                    #ret.add(layer.name, renpy.easy.displayable(img.what), at_list=at_list, key=key, name=img.what.split(' '))
                    what = img.what.split(' ')
                    if what[0]!=img.tag:
                        name = tuple([img.tag,]+what[1:])
                    else:
                        name = tuple(what)
                    ret.add(layer.name, renpy.easy.displayable(img.what), at_list=at_list, key=key, name=name)

                    showing.attributes[(layer.name, key)] = img.what.split(' ')
                elif FUTUREDEBUG:
                    print "Warning: Ignoring SafeloadScreenContext"
        return ret
# add args: (SceneLists, layername, exact_displayable, key=tag
#print "Debug:add(",self,layer,thing,key,zorder,behind,at_list,name,atl
# Debug:add( <renpy.display.core.SceneLists object at 0x373cf90> master <ATLTransform at 7fd7d9a60790> bg 0 [u'vn_overlay'] [<NamedTransform overlay_pos 0x370bf90>] (u'bg', u'rainroad') None


class SafeloadLayerContext(object):
    saybehavior = False
    def __init__(ctx, name):
        ctx.name = name
        ctx.images = [] # Also contains any screens.
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('img '):
            context_stack.append(SafeloadImageContext(line[4:]))
            ctx.images.append(context_stack[-1])
        elif line.startswith('screen '):
            context_stack.append(SafeloadScreenContext(line[7:]))
            ctx.images.append(context_stack[-1])
        elif line=='saybehavior':
            ctx.saybehavior = True
        else:
            raise Exception("Inconsistent save.")

def scrub_plainstr(s):
    for c in s:
        if c not in 'abcdefghijklmnopqrstuvwxyz':
            print "Debug:",s
            raise Exception("Invalid plainstr")
    return s

def scrub_accept_anything(s):
    if INSECURE:
        warnings.append((6,'scrub_accept_anything got',type(s),s))
        return s
    raise Exception("scrub_accept_anything only allowed in insecure mode.")

class SafeloadScreenContext(object):
    ats = None
    name = None
    # XXX this is not a very clean way of determining which varnames are
    # sanitizable.
    ScopeVarScrubbers = () # dict(  # was a dict when I was trying to store
                                    # screen data.
    #    char = scrub_accept_anything,
    #    side_image = scrub_accept_anything,
    #    what = scrub_what,
    #    what_id = scrub_plainstr,
    #    who = scrub_accept_anything,
    #    who_id = scrub_plainstr,
    #    window_id = scrub_plainstr,
    #    )
    def __init__(ctx, name):
        ctx.tag = name
        ctx.scope = dict()
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('name '):
            ctx.name = line[5:]
        elif line.startswith('None '):
            if line[5:] in ctx.ScopeVarScrubbers:
                ctx.scope[line[5:]] = None
            else:
                raise Exception("Not a safe scope variable name: "+line[5:])
        elif line.startswith('str '):
            varname, val = line[4:].split(' ',1)
            if varname in ctx.ScopeVarScrubbers:
                ctx.scope[varname] = ctx.ScopeVarScrubbers[varname](unescape_str(val))
            else:
                raise Exception("Not a safe scope variable name: "+varname)
        else:
            raise Exception("Inconsistent save.")

class SafeloadImageContext(object):
    def __init__(ctx, tag):
        ctx.tag = tag
        ctx.ats = []
    def handle(ctx, context_stack, line):
        if line.startswith('at '):
            ctx.ats = []
            for atname in line[3:].split(','):
                trans = getattr(renpy.store, atname, None)
                if type(trans) is NamedTransform:
                    ctx.ats.append(trans)
                elif trans is None:
                    warnings.append((5,"Unexpected transform name",atname))
                else:
                    warnings.append((5,"Unexpected transform type",type(trans)))
        elif line=='end':
            context_stack.pop()
        elif line.startswith('what '):
            ctx.what = line[5:]
        else:
            raise Exception("Inconsistent save.")

def safeload(lines):
    zero = SafeloadContextZero()
    context_stack = [zero]

    lineno = 0
    for line in lines:
        lineno += 1
        if line.startswith('#'):
            continue
        try:
            context_stack[-1].handle(context_stack, line)
        except Exception,e:
            print "Stack was",context_stack
            print "Line was",lineno,line
            print "Exception was",e
            raise

    sls = renpy.exports.scene_lists()
    for layer in sls.layers:
        sls.layers[layer] = []
        sls.at_list[layer] = dict()

    newlog = renpy.python.RollbackLog()
    for ent in zero.log.entries:
        newent = renpy.python.Rollback()
        newlog.log.append(newent)
        newent.context = renpy.execution.Context(True)
        newent.context.call_stack = list(ent.callstack.stack)
        newent.context.current = ent.current
        newent.context.scene_lists = ent.scene.SceneLists()
        newent.context.images = newent.context.scene_lists.shown
        newent.context.dynamic_stack = ent.dynamicstack.stack
        newent.context.music.update(ent.music_update_data())
        for k,v in ent.stores_update_data().iteritems():
            if k in newent.stores:
                raise Exception("Redefining an existing store.")
            newent.stores[k] = v
    newlog.current = newlog.log[-1]
    roots = zero.log.entries[-1].untainted_roots
    try:
        unroller.last_save_caption = zero.info.caption
    except:
        unroller.last_save_caption = ''
    try:
        unroller.save_started_in_version = zero.info.started_in_version
        unroller.save_played_in_versions = zero.info.played_in_versions
    except:
        unroller.save_started_in_version = 'unknown'
        unroller.save_played_in_versions = ['unknown',]
    currentver = renpy.store.slurpfile('game_version')[0][1]
    if currentver not in unroller.save_played_in_versions:
        unroller.save_played_in_versions.append(currentver)
    if DEBUG:
        # Keep this around for poking.
        global last_zero
        last_zero = zero
    newlog.unfreeze(roots, label='_after_load')
    raise Exception("Unfreeze failed.")

    renpy.exports.restart_interaction()

def asciize(d,recursion=0):
    ret = ''
    if isinstance(d,dict):
        if recursion>3:
            return 'Excess recursion.'
        for k, v in d.iteritems():
            ret += asciize(k,recursion+1)+':'+str(type(v))+'='+asciize(v,recursion+1)+';'
    elif type(d) is bool:
        return str(d)
    else:
        try:
            return str(d)
        except:
            return "<could not asciify "+str(type(d))+">"
    return ret

def escape_str(s):
    ret = '';
    for c in s:
        if c=='\\':
            ret += '\\\\'
        elif c=='\n':
            ret += '\\n'
        else:
            oc = ord(c)
            if oc>=32 and oc<=126:
                ret += c
            elif oc<256:
                ret += "\\x%02x"%oc
            else:
                ret += "\\U%x;"%oc
    return ret

def unescape_str(s):
    ret = ''
    i = 0
    while i<len(s):
        c = s[i]
        if c!='\\':
            ret += c
        else:
            i += 1
            c = s[i]
            if c=='\\':
                ret += '\\'
            elif c=='n':
                ret += '\n'
            elif c=='x':
                ret += chr(int(s[i+1:i+3],16))
                i += 2
            elif c=='U':
                a = ''
                i += 1
                while s[i] in '0123456789abcdef':
                    a += s[i]
                    i += 1
                ret += unichr(int(a,16))
        i += 1
    return ret

def stringify_at_list(at_list):
    ret = []
    for at in at_list:
        if type(at) is NamedTransform:
            ret.append(at.name)
        else:
            print "Warning: Not a named transform:",at
            ret.append(at.varname)
    return ','.join(ret)

def serialize_scene_lists(sls):
    ret = ['scene']
    for layername in sls.layers:
        ret.append('layer '+layername)
        layer = sls.layers[layername]
        ats = sls.at_list[layername]
        for shown in layer:
            if FUTUREDEBUG:
                try:
                    ret.append('#'+repr(shown)+str(type(shown.displayable)))
                except:
                    ret.append('#Unrepresentable'+str(type(shown.displayable)))
            if shown.tag is not None:
                if shown.tag.startswith('hide$'):
                    # So we don't wind up sticking a spurious 'end' after
                    # a hide$ displayable - which would be bad because
                    # it would close the layer.  Also first, because a
                    # hide$ can be a ScreenDisplayable.
                    continue
                elif type(shown.displayable) is renpy.display.screen.ScreenDisplayable:
                    if FUTUREDEBUG: # Screens don't need to be saved in DW.
                        ret.append('screen '+shown.tag)
                        print "Debug: 903",shown.name,shown.tag
                        if ' '.join(shown.name) != shown.tag:
                            ret.append('name '+' '.join(shown.name))
                        for k,v in shown.displayable.scope.iteritems():
                            if type(v) is str or type(v) is unicode:
                                ret.append('str '+k+' '+escape_str(v))
                            elif v is None:
                                ret.append('None '+k)
                            else:
                                ret.append('# Skipped encoding'+k+' '+str(type(v)))
                    else: # To block spurious end.
                        continue
                elif not shown.tag.startswith('hide$'):
                    # Don't save hiding anims.
                    ret.append('img '+shown.tag)
                    if shown.orig_name is not None:
                        if type(shown.orig_name) is str or type(shown.orig_name) is unicode:
                            ret.append('what '+shown.orig_name)
                        else:
                            ret.append('what '+(' '.join(shown.orig_name)))
                    elif shown.name is not None:
                        ret.append(('what '+(' '.join(shown.name))))
                    else:
                        ret.append("#.688")
                    if shown.orig_name is not None and shown.orig_name!=shown.name:
                        ret.append('#.1238 '+(' '.join(shown.name)))
                if shown.tag in ats and ats[shown.tag]:
                    ret.append('at '+stringify_at_list(ats[shown.tag]))
                ret.append('end')
            elif type(shown.displayable) is renpy.display.behavior.SayBehavior:
                ret.append('saybehavior')
            else:
                print "Debug: discarded scene entry with tag None",shown
        ret.append('end')
    ret.append('end')
    ret.append('#/scene')
    return ret

def serialize_datewarp_save_info():
    ret = ['saveinfo']
    try:
        if unroller.last_save_caption:
            ret.append('caption '+escape_str(unroller.last_save_caption))
    except:
        pass
    try:
        if renpy.store.favourite:
            ret.append('favourite '+escape_str(renpy.store.favourite))
    except:
        pass
    try:
        if renpy.store.last_timer:
            ret.append('timer '+escape_str(renpy.store.last_timer))
    except:
        pass
    try:
        ret.append('startver '+unroller.save_started_in_version)
    except:
        ret.append('#startver exception.')
    try:
        ret.append('playvers '+(';'.join(unroller.save_played_in_versions)))
    except:
        ret.append('#playvers exception.')
    ret.append('end')
    return ret

def safesave_serialize():
    return serialize_datewarp_save_info()+serialize_rollback_log(renpy.game.log)

class SaveInfo(object):
    caption = None
    favourite = None
    timer = None
    allowed_favourites = ('Alben','Bradley','Linds','Nathaniel','Rafael')
    mtime = None
    readable = False
    def __init__(self, filename=None):
        if filename:
            self.read_from_file(filename)
    def read_from_file(self, filename):
        try:
            fn = os.path.join(renpy.config.savedir,renpy.store.persistent._file_page+'-'+str(filename)+'.dws')
            st = os.stat(fn)
            self.mtime = st.st_mtime
            f = open(fn, 'rb')
            line = True
            parse = False
            while line:
                line = f.readline()
                if line.endswith('\n'):
                    line = line[:-1]
                if line=='saveinfo':
                    parse = True
                elif not parse:
                    continue
                elif line=='end':
                    self.readable = True
                    return
                else:
                    self.handle([None,],line)
        except Exception,e:
            return
    def handle(self, context_stack, line):
        if line.startswith('caption '):
            # This assumes that caption is used in a context where Ren'Py
            # text tags are interpolated, but [variable] and %(variable)s
            # style tags are not.
            self.caption = unescape_str(line[8:]).replace('{','{{')
        elif line=='end':
            context_stack.pop()
        elif line.startswith('favourite ') and line[10:] in self.allowed_favourites:
            self.favourite = unescape_str(line[10:])
        elif line.startswith('startver '):
            self.started_in_version = line[9:]
        elif line.startswith('playvers '):
            self.played_in_versions = line[9:].split(';')
        elif line.startswith('timer '):
            # This assumes that timer is used in a context where Ren'Py
            # text tags are interpolated, but [variable] and %(variable)s
            # style tags are not.
            self.timer = unescape_str(line[6:]).replace('{','{{')
        else:
            print "Debug: Unexpected savegame info line",line
    def InfoStr(self):
        if self.caption:
            if self.timer:
                return self.caption+'\n'+self.timer
            else:
                return self.caption
        elif self.timer:
            return self.timer
        return ''

autosave_serial = 0
def autosave(savedir):
    global autosave_serial
    while True:
        try:
            f = open(os.path.join(savedir, 'auto', str(autosave_serial)+'.das'),'rb')
            autosave_serial += 1
            f.close()
        except:
            f = open(os.path.join(savedir, 'auto', str(autosave_serial)+'.das'),'wb')
            f.writelines('\n'.join(safesave_serialize()))
            f.close()
            autosave_serial += 1
            return

def try_all_saves(savedir):
    files = os.listdir(savedir)
    warnings.note(("try_all_saves(",savedir,")"))
    for fn in files:
        warnings.note((fn,))
        try:
            if fn.endswith('.das') or fn.endswith('.dws'):
                f = file(os.path.join(savedir,fn),'rb')
                lines = f.read().split('\n')
                f.close()
                safeload(lines)
        except renpy.game.RestartTopContext:
            pass
        except renpy.game.RestartContext:
            pass
        except Exception,e:
            warnings.append((0,"An exception occurred loading",fn,e))
        warnings.unnote()
    warnings.unnote()

def TEST():
    print "Debug: Test cleared",len(warnings)
    clear_warnings()
    try_all_saves(os.path.join(renpy.config.savedir,'auto'))
    print "...  and the warnings went to",len(warnings)
    global uwarnings
    uwarnings = []
    seen = set()
    for warning in warnings:
        try:
            if warning[:-2] not in seen:
                seen.add(warning[:-2])
                uwarnings.append(warning)
        except:
            uwarnings.append(warning)

def init_save_data():
    unroller.last_save_caption = ''
    ver = renpy.store.slurpfile('game_version')[0][1]
    unroller.save_played_in_versions = [ver,]
    partver = renpy.store.slurpfile('partial_update')
    origver = renpy.store.slurpfile('original_game_version')
    if partver:
        ver += ';part:'+partver[0][1]
    if origver:
        ver += ';orig:'+origver[0][1]
    unroller.save_started_in_version = ver

class SafeloadPersistentContext(object):
    preferences = None
    def __init__(self):
        self.data = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif line=='_preferences':
            context_stack.append(SafeloadStoreContext())
            self.preferences = context_stack[-1]
            return
        else:
            self.data.append(parse_tainted_tnv(line, context_stack))
    
class Persistent(object):
    # Built-in values:
    install_beta_versions = None
    install_alpha_versions = None
    _preferences = None

    _ok_underscored = ('_chosen', '_file_page', '_seen_audio', '_seen_ever', '_seen_images', '_set_preferences', '_style_preferences', '_voice_mute')

    # Replacement for the Ren'Py persistent object.
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        print "Returning None for",name
        return None
    def __setattr__(self, name, val):
        if name.startswith('_') and name not in self._ok_underscored:
            warnings.append((5,"Persistent disallowed forbidden preference",name))
        if name in self.__dict__ and self.__dict__[name] == val:
            # Don't bother marking things changed if the change didn't
            # do anything.  XXX this needs to correctly handle member
            # changes.
            return
        self.__dict__[name] = val
    def __getstate__(self):
        print "Persistent.getstate"
        return ()
    def _update(self):
        raise Exception("_update deprecated.")
        if self._preferences is None:
            self.__dict__['_preferences'] = renpy.preferences.Preferences()
        if not self._seen_ever:
            self.__dict__['_seen_ever'] = {}
        if not self._seen_images:
            self.__dict__['_seen_images'] = {}
        if not self._chosen:
            self.__dict__['_chosen'] = {}
        if not self._seen_audio:
            self.__dict__['_seen_audio'] = {}
        if self._changed is None:
            self.__dict__['_changed'] = {}
    def __init__(self):
        self.__dict__['_preferences'] = renpy.preferences.Preferences()
        self.__dict__['_seen_ever'] = {}
        self.__dict__['_seen_images'] = {}
        self.__dict__['_chosen'] = {}
        self.__dict__['_seen_audio'] = {}
        self.__dict__['_changed'] = {}

def persistent_serialize(persistent):
    ret = ['persistent']
    ret.extend(serialize_tnv('install_alpha_versions',persistent.install_alpha_versions))
    ret.extend(serialize_tnv('install_beta_versions',persistent.install_beta_versions))
    for k,v in persistent.__dict__.iteritems():
        if k=='_changed': # This is just a list of the times members were
            # changed last.  It shouldn't be serialized.
            pass
        elif k=='_preferences':
            ret.extend(persistent_serialize_preferences(v))
        else:
            ret.extend(serialize_tnv(k,v))
    ret.append('end')
    return ret
def persistent_serialize_preferences(_preferences):
    ret = ['_preferences']
    for pref,val in _preferences.__dict__.iteritems():
        ret.extend(serialize_tnv(pref, val))
    ret.append('end')
    return ret

def scrub_persistent_changed(s):
    if s:
        warnings.append((5,"scrub_persistent_changed discarded",s))
    return {}

def scrub_datewarp_persistent_chosen(s):
    # Just silently discard the chosen because it's not used in Date Warp.
    # For other games, it will be a dict of (locationtuple):menuoption pairs.
    #if s:
    #    warnings.append((5,'scrub_persistent_chosen discarded',type(s),s))
    return {}

def scrub_persistent_seen_ever(s):
    if type(s) is not dict:
        warnings.append((5,'scrub_persistent_seen_ever discarded',type(s),s))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if v is True:
            if type(k) is str:
                ret[k] = True
            elif type(k) is tuple and len(k)==3 and type(k[0]) is str and type(k[1]) is int and type(k[2]) is int:
                ret[k] = True
            else:
                warnings.append((5,'scrub_persistent_seen_ever discarded',k,v))
        else:
            warnings.append((5,'scrub_persistent_seen_ever discarded',k,v))
    return ret

def ok_persistent_single_seen_image(k,v):
    if v is not True:
        return False
    if type(k) is not tuple:
        return False
    for ik in k:
        if type(ik) is not str:
            return False
    return True

def scrub_persistent_seen_images(s):
    ret = {}
    for k,v in s.iteritems():
        if ok_persistent_single_seen_image(k,v):
            ret[k] = v
        else:
            warnings.append((5,'scrub_persistent_seen_images discarded',k,v))
    return ret

def scrub_persistent_style_preferences(s):
    if s or type(s) is not renpy.python.RevertableDict:
        warnings.append((5,'scrub_persistent_style_preferences discarded',type(s),s))
    return renpy.python.RevertableDict()

def scrub_persistent_voice_mute(s):
    if type(s) is renpy.python.RevertableSet and not s:
        return renpy.python.RevertableSet()
    warnings.append((5,'scrub_persistent_voice_mute discarded',s))
    return renpy.python.RevertableSet()

def scrub_endings_reached(s):
    ret = []
    for k in s:
        if type(k) is int:
            ret.append(k)
        else:
            warnings.append((5,'scrub_endings_reached discarded ending',k))
    return renpy.python.RevertableSet(ret)

def scrub_checklist(s):
    ret = {}
    for k,v in s.iteritems():
        if v is True and type(k) is str:
            ret[k] = True
        else:
            warnings.append((5,'discarded checklist item',k,v))
    return ret

def scrub_wires_cheats(s):
    if type(s) is not dict and type(s) is not renpy.python.RevertableDict:
        warnings.append((5,'scrub_wires_cheats discarded',s))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if (type(v) is not renpy.python.RevertableDict) and (type(v) is not dict):
            warnings.append((5,'scrub_wires_cheats discarded value',type(v),v))
            continue
        if k is None or type(k) is str:
            vret = {}
            ret[k] = vret
            for vk, vv in v.iteritems():
                if type(vk) is str and vv==1:
                    vret[vk] = 1
                else:
                    warnings.append((5,'scrub_wires_cheats discarded item',vk,vv))
        else:
            warnings.append((5,'scrub_wires_cheats discarded',type(k),k,type(v),v))
    return ret

PERSISTENT_SCRUBBERS = dict(
    # Spiky.
    hide_updater = sanitize.expect_bool,
    install_alpha_versions = sanitize.expect_bool_or_None,
    install_beta_versions = sanitize.expect_bool_or_None,
    use_visible_keyboard = sanitize.expect_bool,
    # Datewarp-specific.
    endings_reached = scrub_endings_reached,
    wires_cheats = scrub_wires_cheats,
    checklist = scrub_checklist,
    # Old Ren'Py
    _changed = scrub_persistent_changed,
    _chosen = scrub_datewarp_persistent_chosen,
    _file_page = sanitize.const('1'),
    _invoke_main_menu = sanitize.expect_bool,
    _seen_audio = sanitize.scrub_persistent_seen_audio,
    _seen_ever = scrub_persistent_seen_ever,
    _seen_images = scrub_persistent_seen_images,
    _set_preferences = sanitize.expect_bool,
    _style_preferences = scrub_persistent_style_preferences,
    _voice_mute = scrub_persistent_voice_mute,
    )

PREFERENCE_SCRUBBERS = dict(
    afm_after_click = sanitize.expect_bool,
    afm_enable = sanitize.expect_bool,
    afm_time = sanitize.expect_float_or_int,
    fullscreen = sanitize.expect_bool,
    joymap = sanitize.scrub_preference_joymap,
    language = sanitize.expect_None,
    mouse_move = sanitize.expect_bool,
    mute = sanitize.scrub_preference_mute,
    performance_test = sanitize.expect_bool,
    physical_size = sanitize.scrub_preference_physical_size,
    renderer = sanitize.scrub_preference_renderer,
    show_empty_window = sanitize.expect_bool,
    skip_after_choices = sanitize.expect_bool,
    skip_unseen = sanitize.expect_bool,
    transitions = sanitize.expect_int,
    text_cps = sanitize.expect_int,
    using_afm_enable = sanitize.expect_bool,
    voice_sustain = sanitize.expect_bool,
    volumes = sanitize.scrub_preference_volumes,
    wait_voice = sanitize.expect_bool,
    )
def scrub_persistent_preferences(prefdata):
    ret = renpy.preferences.Preferences()
    for item in prefdata.data:
        if item.name in PREFERENCE_SCRUBBERS:
            setattr(ret, item.name, PREFERENCE_SCRUBBERS[item.name](deindirect_if_needed(item.val)))
        else:
            warnings.append((5,'Discarded unhandled pref',item.name,item.val))
    return ret

def persistent_init():
    try:
        os.stat(os.path.join(renpy.config.savedir,'persistent.dat'))
    except:
        return Persistent()
    return persistent_load(os.path.join(renpy.config.savedir,'persistent.dat'))

def persistent_load(filename):
    if not filename.endswith('.dat'):
        print "Disabled persistent",filename
        return None
    f = open(filename,'rb')
    if f.readline() != 'persistent\n':
        warnings.append((5,"Could not load persistent: bad magic."))
        return
    pctx = SafeloadPersistentContext()
    context_stack = [pctx]
    for line in f.readlines():
        if line.startswith('#'):
            continue
        if line.endswith('\n'):
            line = line[:-1]
        context_stack[-1].handle(context_stack,line)
        if not context_stack:
            break
    ret = Persistent()
    global last_pctx
    last_pctx = pctx
    for item in pctx.data:
        if item.name not in PERSISTENT_SCRUBBERS:
            warnings.append((5,"Discarded unhandled persistent item",item.name,item.val))
        else:
            val = PERSISTENT_SCRUBBERS[item.name](deindirect_if_needed(item.val))
            setattr(ret, item.name, val)
    ret.__dict__['_preferences'] = scrub_persistent_preferences(pctx.preferences)
    return ret

PERSISTENT_BACKUP_COUNT=15
def persistent_cycle():
    i = PERSISTENT_BACKUP_COUNT
    try:
        os.unlink(os.path.join(config.savedir, 'persistent.bak'+str(i)))
    except:
        pass
    while i>0:
        i -= 1
        try:
            os.rename(
                os.path.join(config.savedir, 'persistent.bak'+str(i)),
                os.path.join(config.savedir, 'persistent.bak'+str(i+1)))
        except:
            pass
    try:
        os.rename(
            os.path.join(config.savedir, 'persistent.dat'),
            os.path.join(config.savedir, 'persistent.bak0'))
    except:
        pass

def persistent_write(self, filename):
    f = open(filename+'.new','wb')
    f.write('\n'.join(persistent_serialize(self)))
    f.close()
    try:
        os.unlink(filename)
    except:
        pass
    os.rename(filename+'.new',filename)

def persistent_save():
    if not renpy.game.persistent._changed:
        print "Persistent not changed, saving anyhow?"
        # If nothing's been changed, no need to save anything.
    persistent_cycle()
    try:
        persistent_write(renpy.game.persistent,os.path.join(renpy.config.savedir,'persistent.dat'))
    except Exception,e:
        warnings.append((4,"Error saving persistent.dat",e))
        raise e
